/**
 * Copyright (c) 2010 Keang Ltd
 * 
 * This file is part of the Date Selector package.
 *
 * The Date Selector package is free software: you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as published 
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * The Date Selector package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with the Date Selector package.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.co.keang.date;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Frame;
import java.awt.GridLayout;
import java.awt.HeadlessException;
import java.awt.KeyEventDispatcher;
import java.awt.KeyboardFocusManager;
import java.awt.Window;
import java.awt.Dialog.ModalityType;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.concurrent.Callable;
import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.ActionMap;
import javax.swing.DefaultComboBoxModel;
import javax.swing.InputMap;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JToggleButton;
import javax.swing.KeyStroke;
import javax.swing.SwingConstants;
import javax.swing.Timer;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import org.joda.time.Chronology;
import org.joda.time.DateTime;
import org.joda.time.MutableDateTime;
import org.joda.time.ReadableDateTime;
import uk.co.keang.swingaddons.HoverDetector;
import uk.co.keang.swingaddons.SelectableLineBorder;
import uk.co.keang.swingaddons.VariableGridLayout;


/**
 * An abstract Date Selection GUI.<br>
 * Displays a calendar style view of the days in a month. The month and year can be easily
 * changed via drop down controls and/or by month increment buttons and the day selected by
 * clicking on a day button.
 * <p>
 * If todays date is visible ie the panel is showing the todays month and year then the text
 * on the button with todays day is coloured RED.
 * <p>
 * As well as being able to select a date using the mouse the following key-strokes are recognised:<br>
 * LEFT: move back 1 day<br>
 * RIGHT: move forward one day<br>
 * UP: move back 1 week<br>
 * DOWN: move forward 1 week<br>
 * PAGE_UP: move back 1 month<br>
 * PAGE_DOWN: move forward 1 month<br>
 * HOME: move back 1 year<br>
 * END: move forward 1 year<br>
 * ESC: cancel any selected date and hide the panel<br>
 * ENTER: retain the selected date and hide the panel<br>
 * <p>
 * Listeners can be added to receive notification of a change of selected date and/or the selection of
 * a day button.
 *
 * Factory methods are provided to:<br>
 * 1. Create a dialog to show the selector optionally with OK/Cancel buttons<br>
 * 2. Attach a popup-on-hover feature to a component to popup a date selector if the mouse hovers over
 *    the component.<br>
 * <p>
 * The chronology of the initial date is maintained throughout the life of this object.
 *
 * @author A.G.Docherty
 * @date 14 Jan 2010
 * @version 1.1
 */
/*
 * Version 1.1
 * Pooling of components added to reduce the number of components drawn and discarded
 * 
 */
public abstract class AbstractDateSelector<T> extends JPanel
{
private static final long serialVersionUID = -5551953076052424333L;

protected static final String ACTION_NAME_ENTER = "ENTER";
protected static final String ACTION_NAME_ESC = "ESC";
protected static final String ACTION_NAME_YEAR_INC = "Year+1";
protected static final String ACTION_NAME_YEAR_DEC = "Year-1";
protected static final String ACTION_NAME_MONTH_INC = "Month+1";
protected static final String ACTION_NAME_MONTH_DEC = "Month-1";
protected static final String ACTION_NAME_WEEK_INC = "Day+7";
protected static final String ACTION_NAME_WEEK_DEC = "Day-7";
protected static final String ACTION_NAME_DAY_INC = "Day+1";
protected static final String ACTION_NAME_DAY_DEC = "Day-1";

private static final int MAX_YEARS = 200;
private static final boolean CENTRALISE_DAYS_DISPLAY = true;
private static final int TIME_DELAY_BTN_SELECTED = 500;
private static final long KEY_REPETITION_DELAY = 100;

private static Map<Component, HoverDetector> hoverPopupMap = new HashMap<Component, HoverDetector>();

private int nDayDisplayRows;

private String[] dayNames;
private String[] monthNames;
private Integer[] yearNumbers;

protected Window window;
private JPanel weekNumPanel;
private JPanel daysPanel;
private JPanel dayNamesPanel;
private JButton cancelBtn;
private JButton okBtn;
private JComboBox monthCB;
private JComboBox yearCB;
private DayButton selectedDayButton;
private JButton nextMonthBtn;
private JButton prevMonthBtn;
private JPanel dateSelectionPanel;
private JPanel calPanel;

private DisableCounter layoutDisabled = new DisableCounter();
private DisableCounter eventsDisabled = new DisableCounter();
private boolean showWeekNums = true;
private boolean showDayNames = true;
private boolean showSurroundingDays = true;
private boolean showOkCancelPanel = false;
private DateTime today;

private List<ActionListener> dayListeners = new ArrayList<ActionListener>();
private List<ChangeListener> dateChangeListeners = new ArrayList<ChangeListener>();

protected MutableDateTime baseDate;
protected MutableDateTime movingDate;

protected Chronology chronology;

private DayButton[] daysList;
private JLabel[] weekNumLabels;
private JLabel[] dayLabels;



/**
 * Adds a pop up listener to this component. When the mouse hovers over the component
 * a DateTimeSelector popup will display.
 * The panel will close when:
 * 1. a day button is selected
 * 2. the mouse moves out of the panel
 * 3. the ESCAPE key is pressed
 * 4. the ENTER key is pressed
 * <p>
 * The given listener will be called only if the panel closes due to a day being selected or
 * the ENTER key is pressed
 * <p>
 * @param owner - the component to add the hover popup listener to
 * @param callBack - a callable object that will be called when a day button is selected
 */
static void setPopupOnHoverForComponent(Component owner, Callable callBack)
    {
    HoverDetector popupAction = new HoverDetector(owner, callBack);
    hoverPopupMap.put(owner, popupAction);
    }

/**
 * Removes the Popup on hover functionality from this component
 *
 * @param owner - the component
 */
static void removePopupOnHoverFromComponent(Component owner)
    {
    HoverDetector popupAction = hoverPopupMap.remove(owner);

    if ( popupAction != null )
        {
        popupAction.remove();
        }
    }

/**
 * Displays a modal dialog allowing the selection of a date.
 * This method is provided to allow a Calendar pane to be constructed, configured
 * and then displayed.
 *
 * @param owner - the owning component or null
 * @param title - The title to display in the dialog frame or null for no title 
 * @param panel - the panel
 * @return the selected date or null if cancel pressed
 */
static JDialog createDialog(Component owner, String title, AbstractDateSelector panel)
    {
    JDialog dlg = new JDialog(getWindowForComponent(owner), ModalityType.APPLICATION_MODAL);
    
    if ( title != null )
        dlg.setTitle(title);

    ((AbstractDateSelector<?>)panel).window = dlg;
    dlg.add(panel);

    dlg.pack();

    if ( owner != null )
        dlg.setLocationRelativeTo(owner);
    
    return dlg;
    }

/**
 * Gets the window owner of this component
 *
 * @param parentComponent - the component
 * @return The window containing this component
 * @throws HeadlessException if there is no window owner
 */
static Window getWindowForComponent(Component parentComponent) throws HeadlessException
    {
    if (parentComponent == null)
        return JOptionPane.getRootFrame();

    if (parentComponent instanceof Frame || parentComponent instanceof Dialog)
        return (Window)parentComponent;

    return AbstractDateSelector.getWindowForComponent(parentComponent.getParent());
    }

/**
 * Construct the panel for the default locale
 *
 * @param base - the base date
 * @param end - the end date, this may be before or after the base date 
 * @param includeOkCancel - true to include a panel with OK and Cancel buttons
 */
public AbstractDateSelector(DateTime base, DateTime end, boolean includeOkCancel)
    {
    this(base, end, includeOkCancel, null);
    }

/**
 * Construct the panel
 *
 * @param base - the base date
 * @param end - the end date, this may be before or after the base date 
 * @param includeOkCancel - true to include a panel with OK and Cancel buttons
 * @param loc - the locale to use or null for the default locale
 */
public AbstractDateSelector(DateTime base, DateTime end, boolean includeOkCancel, Locale loc)
    {
    chronology = base.getChronology();
    today = new DateTime(chronology);
    showOkCancelPanel = includeOkCancel;

    initCalendarConstants(loc);
    initYears(base.getYear());
    createDayButtons();
    createSurroundingDayLabels();
    
    baseDate = base.toMutableDateTime(chronology);
    
    if ( base == end )
        movingDate = baseDate;
    else
        movingDate = end.toMutableDateTime(chronology);
    
    selectedDayButton = daysList[movingDate.getDayOfMonth()-1];

    layoutDisabled.reset();
    eventsDisabled.reset();
    
    layoutApplicationPanel(base, includeOkCancel);
    initKeyActions();
    }

/**
 * initialises the constants used for this chronology
 */
private void initCalendarConstants(Locale loc)
    {
    MutableDateTime m = new MutableDateTime(chronology);
    m.setMonthOfYear(1);
    m.setDayOfWeek(1);

    String[] str = new String[m.dayOfWeek().getMaximumValueOverall()];

    for ( int i = 0; i < str.length; i++ )
        {
        str[i] = m.dayOfWeek().getAsShortText(loc);
        m.addDays(1);
        }

    dayNames = str;

    str = new String[m.monthOfYear().getMaximumValueOverall()];
    m.setMonthOfYear(1);

    for ( int i = 0; i < str.length; i++ )
        {
        str[i] = m.monthOfYear().getAsText(loc);
        m.addMonths(1);
        }

    monthNames = str;

    nDayDisplayRows = ((m.dayOfMonth().getMaximumValueOverall() + dayNames.length - 1) / dayNames.length) + 1;
    }

/**
 * Initialises the years to display in the year combo box
 * 
 * @param year - the base year
 */
private void initYears(int year)
    {
    Integer[] years = new Integer[MAX_YEARS];
    int y = year - (years.length/2);
    
    for ( int i = 0; i < years.length; i++ )
        {
        years[i] = Integer.valueOf(y + i);
        }

    yearNumbers = years;
    }


/**
 * init the key strokes that have meaning for this panel
 */
protected void initKeyActions()
    {
    // add the day move key strokes
    InputMap imap = this.getInputMap(WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
    imap.put(KeyStroke.getKeyStroke(KeyEvent.VK_LEFT, 0), ACTION_NAME_DAY_DEC);
    imap.put(KeyStroke.getKeyStroke(KeyEvent.VK_RIGHT, 0), ACTION_NAME_DAY_INC);
    imap.put(KeyStroke.getKeyStroke(KeyEvent.VK_UP, 0), ACTION_NAME_WEEK_DEC);
    imap.put(KeyStroke.getKeyStroke(KeyEvent.VK_DOWN, 0), ACTION_NAME_WEEK_INC);

    ActionMap amap = this.getActionMap();
    amap.put(ACTION_NAME_DAY_INC, new DayChangeAction(1));
    amap.put(ACTION_NAME_DAY_DEC, new DayChangeAction(-1));
    amap.put(ACTION_NAME_WEEK_INC, new DayChangeAction(dayNames.length));
    amap.put(ACTION_NAME_WEEK_DEC, new DayChangeAction(-dayNames.length));

    // add the month move key strokes
    imap.put(KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_UP, 0), ACTION_NAME_MONTH_DEC);
    imap.put(KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_DOWN, 0), ACTION_NAME_MONTH_INC);

    amap.put(ACTION_NAME_MONTH_INC, new MonthChangeAction(1));
    amap.put(ACTION_NAME_MONTH_DEC, new MonthChangeAction(-1));

    // add the year move key strokes
    imap.put(KeyStroke.getKeyStroke(KeyEvent.VK_HOME, 0), ACTION_NAME_YEAR_DEC);
    imap.put(KeyStroke.getKeyStroke(KeyEvent.VK_END, 0), ACTION_NAME_YEAR_INC);

    amap.put(ACTION_NAME_YEAR_INC, new MonthChangeAction(monthNames.length));
    amap.put(ACTION_NAME_YEAR_DEC, new MonthChangeAction(-monthNames.length));

    // add the control keys
    if ( okBtn != null )
        {
        imap.put(KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0), ACTION_NAME_ESC);
        imap.put(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0), ACTION_NAME_ENTER);

        amap.put(ACTION_NAME_ESC, new EscapeAction());
        amap.put(ACTION_NAME_ENTER, new EnterAction());
        }

    // add a key event dispatcher to prevent multiple key events piling up and stalling the GUI
    // when one of the above keys is held down
    KeyboardFocusManager.getCurrentKeyboardFocusManager().addKeyEventDispatcher(new KeyEventDispatcher()
        {
        private long time = 0;
        private int key;
        
        @Override
        public boolean dispatchKeyEvent(KeyEvent e)
            {
            int id = e.getID();
            
            if ( id == KeyEvent.KEY_RELEASED )
                {
                // zero the repetition timer if a key is released
                time = 0;
                }
            else if ( id == KeyEvent.KEY_PRESSED )
                {
                long timeNow = System.currentTimeMillis();
                int keyNow = e.getKeyCode();
                
                // consume repetition events that occur within 'n' millis of the last identical event 
                if ( keyNow == key && timeNow < time + KEY_REPETITION_DELAY )
                    {
                    e.consume();
                    return true;
                    }
                
                time = timeNow;
                key = keyNow;
                }
            
            return false;
            }
        });
    }

/**
 * Create the overall panel
 *
 * @param start - the initial date
 */
private void initApplicationPanel(ReadableDateTime start)
    {
    setLayout(new BorderLayout(0, 5));
    setBorder(new EmptyBorder(5,5,5,5));

    calPanel = new JPanel();
    
    daysPanel = new JPanel();
    daysPanel.setLayout(new GridLayout(0, dayNames.length, 2, 2));

    dateSelectionPanel = initDateSelectionPanel(start);

    dayNamesPanel = initDayNamesPanel();
    weekNumPanel = initWeekNumberPanel();
    }

/**
 * layout the overall panel
 *
 * @param includeOkCancel - true to include the OK Cancel Buttons
 */
private void layoutApplicationPanel(ReadableDateTime start, boolean includeOkCancel)
    {
    removeAll();
    
    initApplicationPanel(start);

    add(dateSelectionPanel, BorderLayout.NORTH);
    add(new JPanel(), BorderLayout.CENTER);

    if ( includeOkCancel )
        {
        JPanel ctrl = initOkCancelPanel();
        add(ctrl, BorderLayout.SOUTH);
        }
    
    layoutCalendarPanel();
    add(calPanel, BorderLayout.CENTER);
    }

/**
 * init the top panel - this holds the date selection controls
 *
 * @param start
 */
private JPanel initDateSelectionPanel(ReadableDateTime start)
    {
    // create the month year select panel
    JPanel top = new JPanel();
    
    // arrow left
    prevMonthBtn = new JButton("<<");
    prevMonthBtn.setFont(getFont());
    prevMonthBtn.addActionListener(new ActionListener()
        {
        @Override
        public void actionPerformed(ActionEvent e)
            {
            layoutDisabled.disable();
            
            boolean single = (e.getModifiers() & ActionEvent.SHIFT_MASK) != ActionEvent.SHIFT_MASK;
            addMonths(single, -1);
    
            layoutDisabled.enable();
            
            layoutCalendarPanel();
            fireDateChangeListener();
            }
        });

    top.add(prevMonthBtn);

    // month
    monthCB = new JComboBox(monthNames);
    monthCB.setFont(getFont());
    monthCB.setSelectedIndex(start.getMonthOfYear()-1);
    monthCB.addActionListener(new ActionListener()
        {
        @Override
        public void actionPerformed(ActionEvent e)
            {
            if ( eventsDisabled.isDisabled() )
                return;
            
            int month = ((JComboBox)e.getSource()).getSelectedIndex()+1;

            // if shift key is pressed then this is the end date else it's a new selection
            if ( (e.getModifiers() & ActionEvent.SHIFT_MASK) == ActionEvent.SHIFT_MASK )
                {
                movingDate.setMonthOfYear(month);
                }
            else
                {
                baseDate.setMonthOfYear(month);
                movingDate.setDate(baseDate);
                }
            
            if ( !layoutDisabled.isDisabled() )
                {
                layoutCalendarPanel();
                fireDateChangeListener();
                }
            }
        });

    top.add(monthCB);

    // year
    yearCB = new JComboBox(yearNumbers);
    yearCB.setFont(getFont());
    yearCB.setSelectedItem(Integer.valueOf(start.getYear()));
    yearCB.setEditable(true);
    yearCB.addActionListener(new ActionListener()
        {
        @Override
        public void actionPerformed(ActionEvent e)
            {
            if ( eventsDisabled.isDisabled() )
                return;
            
            Integer ny = (Integer)((JComboBox)e.getSource()).getSelectedItem();
            int y = ny.intValue();
            
            if ( y < yearNumbers[0] + 10 || y > yearNumbers[yearNumbers.length-1] - 10 )
                {
                initYears(y);
                DefaultComboBoxModel m = (DefaultComboBoxModel)yearCB.getModel();
                m.removeAllElements();
                for ( int i = 0; i < yearNumbers.length; i++ )
                    {
                    m.addElement(yearNumbers[i]);
                    }
                
                yearCB.setSelectedItem(ny);
                return;
                }
                
            // if shift key is pressed then this is the end date else it's a new selection
            if ( (e.getModifiers() & ActionEvent.SHIFT_MASK) == ActionEvent.SHIFT_MASK )
                {
                movingDate.setYear(y);
                }
            else
                {
                baseDate.setYear(y);
                movingDate.setDate(baseDate);
                }

            if ( !layoutDisabled.isDisabled() )
                {
                layoutCalendarPanel();
                fireDateChangeListener();
                }
            }
        });

    top.add(yearCB);

    // arrow right
    nextMonthBtn = new JButton(">>");
    nextMonthBtn.setFont(getFont());
    nextMonthBtn.addActionListener(new ActionListener()
        {
        @Override
        public void actionPerformed(ActionEvent e)
            {
            layoutDisabled.disable();
            
            boolean single = (e.getModifiers() & ActionEvent.SHIFT_MASK) != ActionEvent.SHIFT_MASK;
            addMonths(single, 1);
            
            layoutDisabled.enable();
            
            layoutCalendarPanel();
            fireDateChangeListener();
            }
        });

    top.add(nextMonthBtn);

    return top;
    }

/**
 * Initialise a JPanel for displaying the short names of the days in a week ie Mon to Sun
 * 
 * @return the panel populated with the names
 */
private JPanel initDayNamesPanel()
    {
    // create the day names
    int nDaysInWeek = dayNames.length;
    JPanel p = new JPanel();
    p.setLayout(new GridLayout(1, nDaysInWeek));
    p.setBorder(new CompoundBorder(new EmptyBorder(0, 0, 3, 0),
                   new SelectableLineBorder(EnumSet.of(SelectableLineBorder.Side.BOTTOM, SelectableLineBorder.Side.TOP), Color.BLACK, 1)));

    for ( int i = 0; i < dayNames.length; i++ )
        {
        JLabel lbl = new JLabel(dayNames[i], SwingConstants.CENTER);
        lbl.setFont(getFont());

        p.add(lbl);
        }

    return p;
    }

/**
 * Layout the calendar panel - this panel has the names of the day along
 * the top of the panel, the week numbers along left hand side and the days of the month
 * in remaining section of the panel
 *
 * @return the panel
 */
private void layoutCalendarPanel()
    {
    if ( layoutDisabled.isDisabled() )
        return;
    
    calPanel.setLayout(new VariableGridLayout(showDayNames?2:1,0));
    
    calPanel.removeAll();
    
    layoutDaysDisplay();
    
    // create the centre section which comprises of a day name heading and the calendar

    if ( showWeekNums && showDayNames )
        {
        // this is the empty corner component 
        calPanel.add(new JLabel());
        }

    if ( showDayNames )
        {
        calPanel.add(dayNamesPanel);
        }

    if ( showWeekNums )
        {
        updateWeekNumberPanel();
        
        calPanel.add(weekNumPanel);
        }

    calPanel.add(daysPanel);
    
    revalidate();
    }

/**
 * Creates a button for each day of the longest month in a year
 * 
 */
private void createDayButtons()
    {
    // create the calendar panel buttons action listener to fire an event when a day button
    // is selected. There is a delay on firing the events to prevent multiple events being fired
    // if a number of buttons are selected in quick succession eg if the keys are used to move to a 
    // new day crossing through many other day buttons
    ActionListener listener = new ActionListener()
        {
        private Timer t;
        private ActionEvent event;

        {
        // create a timer to stop events being fired if the keys are used 
        // to navigate across multiple days
        t = new Timer(TIME_DELAY_BTN_SELECTED, new ActionListener()
            {
            @Override
            public void actionPerformed(ActionEvent e)
                {
                fireDayBtnListener(event);
                fireDateChangeListener();
                }
            });

        t.setRepeats(false);
        }
        
        @Override
        public void actionPerformed(ActionEvent e)
            {
            t.restart();
            event = e;
            
            selectedDayButton = (DayButton)e.getSource();

            // if shift key is pressed then this is the end date else it's a new selection
            if ( (e.getModifiers() & ActionEvent.SHIFT_MASK) == ActionEvent.SHIFT_MASK )
                {
                movingDate.setDayOfMonth(selectedDayButton.getDay());
                }
            else
                {
                baseDate.setDayOfMonth(selectedDayButton.getDay());
                movingDate.setDate(baseDate);
                }

            selectBtns();
            }
        };
        
    // create a button of each day of the longest month
    int maxdays = today.dayOfMonth().getMaximumValueOverall();
    daysList = new DayButton[maxdays];

    for ( int i = 0; i < maxdays; i++ )
        {
        DayButton daybtn = new DayButton(i+1);
        daybtn.addActionListener(listener);
        daysList[i] = daybtn;
        }
    }

/**
 * Creates a button for each day of the longest month in a year
 * 
 */
private void createSurroundingDayLabels()
    {
    // create a label of each surrounding day
    int mindays = today.dayOfMonth().getMinimumValueOverall();
    int nLbls = (nDayDisplayRows * dayNames.length) - mindays;
    
    dayLabels = new JLabel[nLbls];
    for ( int i = 0; i < nLbls; i++ )
        {
        dayLabels[i] = new JLabel("", SwingConstants.CENTER);
        }
    }

/**
 * Initialise a JPanel for displaying week numbers
 */
private JPanel initWeekNumberPanel()
    {
    JPanel p = new JPanel();
    p.setFont(getFont());
    p.setLayout(new GridLayout(0, 1, 0, 2));
    p.setBorder(new CompoundBorder(new EmptyBorder(0, 0, 0, 3),
                           new SelectableLineBorder(EnumSet.of(SelectableLineBorder.Side.RIGHT, SelectableLineBorder.Side.LEFT), Color.BLACK, 1)));

    weekNumLabels = new JLabel[nDayDisplayRows];
    
    // add the week number label for each row that contains days
    for ( int i = 0; i < nDayDisplayRows; i++ )
        {
        JLabel lbl = new JLabel("", SwingConstants.CENTER);
        lbl.setFont(getFont());
        p.add(lbl);
        
        weekNumLabels[i] = lbl;
        }
    
    return p;
    }

/**
 * Initialise a JPanel for displaying week numbers
 */
private void updateWeekNumberPanel()
    {
    MutableDateTime date = movingDate;
    
    // get the offset of first day of the month
    int offset = startOfMonthWeekday(date);

    int wkRow1 = date.getWeekOfWeekyear() - ((offset + date.getDayOfMonth()-1) / dayNames.length);

    // the max week number depends on if we are in January (uses previous year value) or
    // December (uses this years value).
    int maxNumWeeks;

    if ( date.getMonthOfYear() == 1 )
        {
        date.addYears(-1);
        maxNumWeeks = date.weekOfWeekyear().getMaximumValue();
        date.addYears(1);
        }
    else
        {
        maxNumWeeks = date.weekOfWeekyear().getMaximumValue();
        }

    // add the trailing blanks to ensure the number of displayed rows is the same for each month
    // get the maximum num of days for this month
    int maxDays = date.dayOfMonth().getMaximumValue();

    int extraLabels = 0;
    
    if ( !showSurroundingDays )
        {
        extraLabels = (nDayDisplayRows * dayNames.length) - (dayNames.length - 1) - offset - maxDays;
        }
    
    JLabel lbl = new JLabel("8888");
    lbl.setFont(getFont());
    Dimension d = lbl.getPreferredSize();

    // add the week number label for each row that contains days
    for ( int i = 0; i < weekNumLabels.length; i++ )
        {
        String text;
        
        // don't show a week number if there are extra labels (labels at after the last day of the month) and this is the last row or
        // there is more than 1 row of extra labels and this is the last but one row or
        // we are not displaying surrounding days and the first day is on the second row and this is the first row
        if ( (extraLabels > 0 && i == nDayDisplayRows - 1) || 
             (extraLabels > dayNames.length && i == nDayDisplayRows - 2) ||
             (!showSurroundingDays && offset >= dayNames.length && i == 0) )
            {
            text = "";
            }
        else
            {
            int wk = wkRow1 + i;

            // if the wk is greater than maxNumWeeks then its the start of the new year
            // if it's less than 1 then it's from the previous year
            if ( wk > maxNumWeeks )
                wk -= maxNumWeeks;
            else if ( wk < 1 )
                wk += maxNumWeeks;

            text = String.valueOf(wk);
            }

        lbl = weekNumLabels[i];
        lbl.setText(text);
        lbl.setFont(getFont());
        lbl.setPreferredSize(d);
        weekNumPanel.add(lbl);
        }
    }

/**
 * create the OK cancel control panel
 */
private JPanel initOkCancelPanel()
    {
    // create the ok cancel panel
    JPanel bot = new JPanel();

    // add the OK Button
    okBtn = new JButton();
    okBtn.setFont(getFont());
    okBtn.setAction(new EnterAction());
    bot.add(okBtn);


    // add the Cancel Button
    cancelBtn = new JButton();
    cancelBtn.setFont(getFont());
    cancelBtn.setAction(new EscapeAction());
    bot.add(cancelBtn);


    return bot;
    }

/**
 * Change the month by the given offset
 *
 * @param offset - the amount to change - negative for earlier dates, positive for later dates
 */
private void addMonths(boolean single, int offset)
    {
    MutableDateTime date;
    
    if ( single )
        {
        date = baseDate;
        }
    else
        {
        date = movingDate;
        }

    date.addMonths(offset);
    int m = date.getMonthOfYear();

    while ( m > monthNames.length )
        {
        // roll round a year
        m -= monthNames.length;
        }

    
    while ( m < 1 )
        {
        // roll round a year
        m += monthNames.length;
        }

    layoutDisabled.disable();
    
    yearCB.setSelectedItem(Integer.valueOf(date.getYear()));
    monthCB.setSelectedIndex(m-1);
    layoutDisabled.enable();
    }

/**
 * Lays out the calendar days display for the current year, month and day
 * If the current day is no longer valid ie the current day is the 31st and the
 * calendar has change to a 30 day month the nearest lower date will become the currently
 * selected day
 *
 */
private JPanel layoutDaysDisplay()
    {
    daysPanel.removeAll();
    
    MutableDateTime date = movingDate;
    
    // get the offset of first day of the month
    int offset = startOfMonthWeekday(date);

    // get the maximum num of days for this month
    int maxDays = date.dayOfMonth().getMaximumValue();

    // change the selectedDay to the display the current values
    selectBtns();
    
    Dimension daysPreferredSize = daysList[0].getPreferredSize();
    int lblIndex = 0;
    
    // add the leading blanks
    if ( showSurroundingDays )
        {
        int max = date.toDateTime(date.getChronology()).minusMonths(1).dayOfMonth().getMaximumValue();

        for ( int i = offset-1; i >= 0; i-- )
            {
            JLabel lbl = dayLabels[lblIndex++];
            lbl.setText(String.valueOf(max-i));
            lbl.setFont(getFont());
            lbl.setPreferredSize(daysPreferredSize);
            daysPanel.add(lbl);
            }
        }
    else
        {
        for ( int i = 0; i < offset; i++ )
            {
            JLabel lbl = dayLabels[lblIndex++];
            lbl.setText("");
            lbl.setFont(getFont());
            lbl.setPreferredSize(daysPreferredSize);
            daysPanel.add(lbl);
            }
        }
    
    // add the days highlighting today's day if it is displayed
    boolean todayDisplayed = ( date.getYear() == today.getYear() && date.getMonthOfYear() == today.getMonthOfYear() );

    int todayDay = today.getDayOfMonth();

    for ( int i = 0; i < maxDays; i++ )
        {
        DayButton btn = daysList[i];

        // set a different text colour for today
        if ( todayDisplayed && todayDay == (i+1) )
            btn.setForeground(Color.RED);
        else
            btn.setForeground(Color.BLACK);

        daysPanel.add(btn);
        }

    // add the extra trailing labels
    int extraLabels = (nDayDisplayRows * dayNames.length) - offset - maxDays;

    for ( int i = 0; i < extraLabels; i++ )
        {
        JLabel lbl = dayLabels[lblIndex++];
        
        if ( showSurroundingDays )
            lbl.setText(String.valueOf(i+1));
        else
            lbl.setText("");
        lbl.setFont(getFont());
        lbl.setPreferredSize(daysPreferredSize);
        daysPanel.add(lbl);
        }
    
    selectedDayButton.requestFocus();
    
    return daysPanel;
    }

/**
 * Selects all buttons
 */
private void selectBtns()
    {
    int ds, de;
    
    if ( ( baseDate.getYear() == movingDate.getYear() && baseDate.getMonthOfYear() == movingDate.getMonthOfYear() ) )
        {
        // both the base and moving dates are displayed
        if ( baseDate.isBefore(movingDate) )
            {
            ds = baseDate.getDayOfMonth()-1;
            de = movingDate.getDayOfMonth()-1;
            }
        else
            {
            ds = movingDate.getDayOfMonth()-1;
            de = baseDate.getDayOfMonth()-1;
            }
        }
    else
        {
        // only the moving date is visible
        if ( baseDate.isBefore(movingDate) )
            {
            ds = 0;
            de = movingDate.getDayOfMonth()-1;
            }
        else
            {
            ds = movingDate.getDayOfMonth()-1;
            de = movingDate.dayOfMonth().getMaximumValue()-1;
            }
        }

    // de-select the range of buttons
    for ( int i = 0; i < ds; i++ )
        {
        daysList[i].setSelected(false);
        }
    
    // select the range of buttons
    for ( int i = ds; i <= de; i++ )
        {
        daysList[i].setSelected(true);
        }
    
    // de-select the range of buttons
    for ( int i = de+1; i < daysList.length; i++ )
        {
        daysList[i].setSelected(false);
        }
    }
/**
 * Gets the 0 based weekday for the first day of this month
 * 
 * @return the 0 based weekday of the first day of the month
 */
private int startOfMonthWeekday(MutableDateTime date)
    {
    int currentDay = date.getDayOfMonth();
    int ndays = dayNames.length;
    int offset = date.getDayOfWeek() - (currentDay%ndays);

    // make sure the numbers start as close to the top left as possible
    if ( offset < 0 )
        offset += ndays;
    else if ( offset >= ndays )
        offset -= ndays;
    
    if ( CENTRALISE_DAYS_DISPLAY )
        {
        // if there is more than 1 empty row 
        // attempt to centralise the display so the other month days don't all come at the end
        if ( (nDayDisplayRows * ndays)-date.dayOfMonth().getMaximumValue() > ndays )
            {
            if ( offset < (ndays/2) )
                offset += ndays;
            }
        }
    
    return offset;
    }

/**
 * Gets the selected date
 *
 * @return the selected date or null if cancel was pressed
 */
public abstract T getSelectedDate() ;

/**
 * Sets the selected date.
 * Once the values have been set the overriding method should call updateOnSettingSelectedDate()
 * to update the display.
 *
 * @param date - the selected date
 */
public abstract void setSelectedDate(T date) ;


protected void updateOnSettingSelectedDate()
    {
    eventsDisabled.disable();
    addMonths(baseDate.isEqual(movingDate), 0);
    eventsDisabled.enable();
    
    layoutCalendarPanel();
    repaint();
    fireDateChangeListener();
    }


/**
 * Gets the Show Week Numbers state
 *
 * @return true if the panel will display week numbers
 */
public boolean isShowWeekNums()
    {
    return showWeekNums;
    }

/**
 * Sets the Show Week Numbers state
 *
 * @param show - true to display week numbers, false to hide them
 */
public void setShowWeekNums(boolean show)
    {
    if ( showWeekNums != show )
        {
        showWeekNums = show;
        layoutCalendarPanel();
        }
    }

/**
 * Gets the Show Day Names display state
 *
 * @return true if the panel will display Day Names
 */
public boolean isShowDayNames()
    {
    return showDayNames;
    }

/**
 * Sets the Show Day Names display state
 *
 * @param show - true to display Day Names, false to hide them
 */
public void setShowDayNames(boolean show)
    {
    if ( showDayNames != show )
        {
        showDayNames = show;
        layoutCalendarPanel();
        }
    }

/**
 * Gets the Show Surrounding Days display state
 *
 * @return true if the panel will display the surrounding days from neighbouring months
 */
public boolean isShowSurroundingDays()
    {
    return showSurroundingDays;
    }

/**
 * Sets the Show Surrounding Days display state
 *
 * @param show - true to display Surrounding Days, false to hide them
 */
public void setShowSurroundingDays(boolean show)
    {
    if ( showSurroundingDays != show )
        {
        showSurroundingDays = show;
        layoutCalendarPanel();
        repaint();
        }
    }

/**
 * Gets the Show OK Cancel Panel
 *
 * @return true if the panel will display OK & Cancel Panel
 */
public boolean isShowOkCancelPanel()
    {
    return showOkCancelPanel;
    }

/**
 * Sets the Show Ok Cancel Panel state
 *
 * @param show - true to display OK Cancel Panel
 */
public void setShowOkCancelPanel(boolean show)
    {
    if ( showOkCancelPanel != show )
        {
        showOkCancelPanel = show;
        layoutCalendarPanel();
        repaint();
        }
    }

/**
 * Adds an action listener to each of the day buttons. This listener
 * will only be called when a DayButton is pressed. Changes to the month or year will not
 * cause this event to be fired.
 *
 * @param l - the listener to add
 */
public void addDayActionListener(ActionListener l)
    {
    dayListeners.add(l);
    }

/**
 * Removes the action listener from each of the day buttons.
 *
 * @param l - the listener to remove
 */
public void removeDayActionListener(ActionListener l)
    {
    dayListeners.remove(l);
    }

/**
 * Adds an date change listener. This listener will be called whenever the date changes.
 * This can be through selecting a new day button, changing the month or the changing the year.
 *
 * @param l - the listener to add
 */
public void addDateChangedListener(ChangeListener l)
    {
    dateChangeListeners.add(l);
    }

/**
 * Removes the date change listener
 *
 * @param l - the listener to remove
 */
public void removeDateChangedListener(ChangeListener l)
    {
    dateChangeListeners.remove(l);
    }

/**
 * fires the date change event to all listeners
 */
protected void fireDateChangeListener()
    {
    ChangeEvent e = new ChangeEvent(this);
    
    // call all listeners
    for ( ChangeListener dl : dateChangeListeners )
        {
        dl.stateChanged(e);
        }
    }

/**
 * fires the day selected event to all listeners
 *
 * @param e - the event to fire
 */
protected void fireDayBtnListener(ActionEvent e)
    {
    // call all day button listeners
    for ( ActionListener al : dayListeners )
        {
        al.actionPerformed(e);
        }
    }

/**
 * Adds the number of days
 *
 * @param nDays - the number of days to add
 */
private void addDays(boolean single, int ndays)
    {
    if ( selectedDayButton == null )
        return;

    MutableDateTime date;
    
    // if this is a new selection use the base time 
    if ( single )
        {
        date = baseDate;
        }
    else
        {
        date = movingDate;
        }
    
    int index = 0;
    int day = selectedDayButton.getDay();

    // if this in still in this month
    if ( day + ndays <= date.dayOfMonth().getMaximumValue() && day + ndays > 0 )
        {
        index = day + ndays - 1;
        }
    else
        {
        date.addDays(ndays);
        index = date.getDayOfMonth() - 1;

        // select new month
        addMonths(single, 0);
        layoutCalendarPanel();
        }

    selectedDayButton = daysList[index];
    selectedDayButton.doClick();
    }

/**
 * Sets the Font for all the components.
 * After changing the font size you may need to call pack() to resize the dialog
 *
 * @param f - the new font
 */
@Override
public void setFont(Font f)
    {
    super.setFont(f);

    if ( daysList != null )
        {
        for ( DayButton btn : daysList )
            {
            btn.setFont(f);
            }

        nextMonthBtn.setFont(f);
        prevMonthBtn.setFont(f);

        monthCB.setFont(f);
        yearCB.setFont(f);

        for ( Component c : dayNamesPanel.getComponents() )
            {
            c.setFont(f);
            }
        
        for ( Component c : weekNumPanel.getComponents() )
            {
            c.setFont(f);
            }
        
        layoutApplicationPanel(baseDate, showOkCancelPanel);
        }
    
    }

/**
 * Sets the action for the OK button
 * 
 * @param a - the action to use
 */
public void setOkBtnAction(Action a) { okBtn.setAction(a); }

/**
 * Sets the action for the Cancel button
 * 
 * @param a - the action to use
 */
public void setCancelBtnAction(Action a) { cancelBtn.setAction(a); }

/**
 * JToggleButton that knows what day it represents
 *
 * @author A.G.Docherty
 * @date 15 Jan 2010
 *
 */
public static class DayButton extends JToggleButton
    {
    private static final long serialVersionUID = 4399751865083856271L;
    private int day;

    DayButton(int d)
        {
        super(String.valueOf(d));
        day = d;
        }

    public int getDay() { return day; }
    
    public String toString()
        {
        return "Day "+day+": "+ super.toString();
        }
    }

/**
 * Called when the day needs to be changed
 *
 * @author A.G.Docherty
 * @date 18 Jan 2010
 *
 */
private class DayChangeAction extends AbstractAction
    {
    private static final long serialVersionUID = 1L;
    int ndays = 0;

    DayChangeAction(int days)
        {
        ndays = days;
        }

    @Override
    public void actionPerformed(ActionEvent e)
        {
        // if shift key is pressed then this is the end date else it's the start date
        boolean start = (e.getModifiers() & ActionEvent.SHIFT_MASK) != ActionEvent.SHIFT_MASK;

        addDays(start, ndays);
        }
    }

/**
 * Called when the month needs to be changed
 *
 * @author A.G.Docherty
 * @date 18 Jan 2010
 *
 */
private class MonthChangeAction extends AbstractAction
    {
    private static final long serialVersionUID = 1L;
    int nMonths = 0;

    MonthChangeAction(int months)
        {
        nMonths = months;
        }

    @Override
    public void actionPerformed(ActionEvent e)
        {
        // if shift key is pressed then this is the end date else it's the start date
        boolean start = (e.getModifiers() & ActionEvent.SHIFT_MASK) != ActionEvent.SHIFT_MASK;
            
        addMonths(start, nMonths);

        layoutCalendarPanel();
        fireDateChangeListener();
        }
    }

/**
 * The Action to use for the Enter key
 *
 * @author A.G.Docherty
 * @date 22 Jan 2010
 *
 */
private class EnterAction extends AbstractAction
    {
    private static final long serialVersionUID = 8473847282367196934L;

    EnterAction()
        {
        putValue(NAME, "OK");
        }

    
    @Override
    public void actionPerformed(ActionEvent e)
        {
        window.setVisible(false);
        window.dispose();
        fireDateChangeListener();
        }
    }

/**
 * The Action to use for the Escape key
 *
 * @author A.G.Docherty
 * @date 22 Jan 2010
 *
 */
private class EscapeAction extends AbstractAction
    {
    private static final long serialVersionUID = 124723384721294196L;

    EscapeAction()
        {
        putValue(NAME, "Cancel");
        }
    
    @Override
    public void actionPerformed(ActionEvent e)
        {
        baseDate = null;
        window.setVisible(false);
        window.dispose();
        }
    }



private class DisableCounter
    {
    int count = 0;
    void disable() { count++; }
    void enable() { count--; }
    void reset() { count = 0; }
    boolean isDisabled() { return count > 0; }
    }

}
