/** 
 * Copyright (c) 2010 Keang Ltd
 * 
 * This file is part of the Date Selector package.
 *
 * The Date Selector package is free software: you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as published 
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * The Date Selector package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with the Date Selector package.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.co.keang.date;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Event;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import java.util.List;
import javax.swing.AbstractAction;
import javax.swing.ActionMap;
import javax.swing.InputMap;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.KeyStroke;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.LineBorder;

/**
 * A dialog for displaying implementations of the AbstractDateSelector in a modal 
 * undecorated panel. The panel will display in a location relative to the owning
 * component.
 * <p>
 * The panel registers additional key strokes for the Escape and Enter keys to
 * close the panel and cancel/accept the selection.
 * <p>
 * If the panel does not display OK/Cancel buttons the panel will close automatically close, 
 * accepting the selection, if a new date is selected.
 * <p>
 * The panel will automatically close, cancelling the selection, if
 * the focus is lost or the mouse exits the panel.  
 * <p>
 * Concrete implementations of this class must implement the getSelector() and getSelectedDate()
 * methods.
 *
 * @author A.G.Docherty
 * @date 20 Jan 2010
 *
 */
abstract class AbstractPopupSelector<P extends AbstractDateSelector<V>,V> extends JDialog
{
private static final long serialVersionUID = 2594335458785943032L;
protected P panel;
private Component owner;

private List<ActionListener> actionListeners = new ArrayList<ActionListener>();

/**
 * Creates a Popup Selector for this component which displays the Date Selector panel
 *
 * @param parent - the component that triggered the popup
 * @param pane - the selector panel to display
 */
public AbstractPopupSelector(Component parent, P pane)
    {
    super(AbstractDateSelector.getWindowForComponent(parent), ModalityType.MODELESS);
    setUndecorated(true);

    panel = pane;
    owner = parent;
    panel.window = this;
    panel.setBorder(new CompoundBorder(new LineBorder(Color.DARK_GRAY), new EmptyBorder(3,3,3,3)));
    add(panel);

    pack();

    if ( owner != null )
        setLocationRelativeTo(owner);

    panel.addMouseListener(new MouseAdapter()
        {
        @Override
        public void mouseExited(MouseEvent e)
            {
            Dimension size = panel.getSize();
            Point p = e.getPoint();

            if ( p.x < 0 || p.y < 0 || p.x > size.width || p.y > size.height )
                {
                // exit without firing a change event
                setVisible(false);
                dispose();
                }
            }
        });
    
    if ( !panel.isShowOkCancelPanel() )
        {
        panel.addDayActionListener(new ActionListener()
            {
            @Override
            public void actionPerformed(ActionEvent e)
                {
                fireActionEvent();
                setVisible(false);
                dispose();
                }
            });
        }
    else
        {
        panel.setOkBtnAction(new EnterAction());
        }
    
    initKeyActions();

    setVisible(true);
    }

/**
 * Initialise the key stroke actions
 */
private void initKeyActions()
    {
    // add the day move key strokes
    InputMap imap = panel.getInputMap(JPanel.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
    ActionMap amap = panel.getActionMap();
    // add the control keys
    imap.put(KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0), "POPUP_ESC");
    imap.put(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0), "POPUP_ENTER");

    amap.put("POPUP_ESC", new AbstractAction()
        {
        private static final long serialVersionUID = -3120863051757803275L;

        @Override
        public void actionPerformed(ActionEvent e)
            {
            // exit without firing a change event
            setVisible(false);
            dispose();
            }
        });

    amap.put("POPUP_ENTER", new AbstractAction()
        {
        private static final long serialVersionUID = 1610619687910267935L;

        @Override
        public void actionPerformed(ActionEvent e)
            {
            fireActionEvent();
            setVisible(false);
            dispose();
            }
        });
    }

/**
 * Adds a listener to this popup to receive notification when a date is selected
 *
 * @param l - the listener
 */
public void addActionListener(ActionListener l)
    {
    actionListeners.add(l);
    }

/**
 * removes a listener from this popup
 *
 * @param l - the listener
 */
public void removeActionListener(ActionListener l)
    {
    actionListeners.remove(l);
    }

/**
 * fires a notification to all registered listeners that a date has been selected and the popup closed
 *
 */
protected void fireActionEvent()
    {
    ActionEvent event = new ActionEvent(panel, Event.ACTION_EVENT, "New Date");

    for ( ActionListener listener : actionListeners )
        {
        listener.actionPerformed(event);
        }
    }

/**
 * Gets the Date Selector panel
 *
 * @return the panel
 */
public abstract P getSelector() ;

/**
 * Gets the selected date from the Date Selector
 * This is a convenience method that calls through to the Date Selector panel
 *
 * @return the selected date
 */
public abstract V getSelectedDate() ;

/**
 * The Action to use for the Enter key
 *
 * @author A.G.Docherty
 * @date 22 Jan 2010
 *
 */
private class EnterAction extends AbstractAction
    {
    private static final long serialVersionUID = 8473847282367196934L;

    EnterAction()
        {
        putValue(NAME, "OK");
        }

    
    @Override
    public void actionPerformed(ActionEvent e)
        {
        fireActionEvent();
        setVisible(false);
        dispose();
        }
    }

}
