/** 
 * Copyright (c) 2010 Keang Ltd
 * 
 * This file is part of the Date Selector package.
 *
 * The Date Selector package is free software: you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as published 
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * The Date Selector package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with the Date Selector package.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.co.keang.date;

import java.awt.Component;
import java.awt.event.ActionListener;
import java.util.Locale;
import java.util.concurrent.Callable;
import javax.swing.JDialog;
import org.joda.time.DateTime;
import org.joda.time.MutableDateTime;


/**
 * A Date selection GUI.<br>
 * Displays a calendar style view of the days in a month. The month and year can be easily
 * changed via drop down controls and/or by month increment buttons and the day selected by
 * clicking on a day button.
 * <p>
 * If todays date is visible ie the panel is showing the todays month and year then the text
 * on the button with todays day is coloured RED.
 *<p>
 * As well as being able to select a date using the mouse the following key-strokes are recognised:<br>
 * LEFT: move back 1 day<br>
 * RIGHT: move forward one day<br>
 * UP: move back 1 week<br>
 * DOWN: move forward 1 week<br>
 * PAGE_UP: move back 1 month<br>
 * PAGE_DOWN: move forward 1 month<br>
 * HOME: move back 1 year<br>
 * END: move forward 1 year<br>
 * ESC: cancel any selected date and hide the panel<br>
 * ENTER: retain the selected date and hide the panel<br>
 * <p>
 * Listeners can be added to receive notification of a change of selected date and/or the selection of
 * a day button.
 * <p>
 * Factory methods are provided to:<br>
 * 1. Show the selector in a JDialog optionally with OK/Cancel buttons<br>
 * 2. Show the selector in a Popup style panel<br>
 * 3. Attach a popup-on-hover feature to a component to popup a date selector if the mouse hovers over
 *    the component.<br>
 * <p>
 * Alternatively the panel can be constructed directly and added to an existing GUI.
 *
 * @author A.G.Docherty
 * @date 14 Jan 2010
 *
 */
public class DateSelector extends AbstractDateSelector<DateTime>
{
private static final long serialVersionUID = -1157816398877047100L;

/**
 * Adds a pop up listener to this component. When the mouse hovers over the component
 * a DateSelector popup will display.
 * <p>
 * The panel will close when:
 * 1. a day button is selected
 * 2. the mouse moves out of the panel
 * 3. the ESCAPE key is pressed
 * 4. the ENTER key is pressed
 * <p>
 * The given listener will be called only if the panel closes due to a day being selected or
 * the ENTER key being pressed
 *
 * @param owner - the component to add the hover popup listener to
 * @param dt - the date that will display when the popup pops up. This objects date can be dynamically 
 *             changed by the calling code to change the date that displays on popup
 * @param l - a listener that will be called when a day button is selected
 */
public static void setPopupOnHoverForComponent(Component owner, MutableDateTime dt, ActionListener l)
    {
    AbstractDateSelector.setPopupOnHoverForComponent(owner, new HoverAction(owner, dt, l));
    }

/**
 * Removes the Popup on hover functionality from this component
 *
 * @param owner - the component
 */
public static void removePopupOnHoverFromComponent(Component owner)
    {
    AbstractDateSelector.removePopupOnHoverFromComponent(owner);
    }

/**
 * Creates a popup panel allowing the selection of a date. The panel is an undecorated
 * JDialog. To show/hide the panel use setVisible(boolean);
 *
 * @param owner - the owning component or null
 * @param now - the time now
 * @return the popup panel
 */
public static PopupDateSelector createPopup(Component owner, DateTime now)
    {
    DateSelector panel = new DateSelector(now, false);
    return createPopup(owner, panel);
    }

/**
 * Creates a popup panel allowing the selection of a date. The panel is an undecorated
 * JDialog. To show/hide the panel use setVisible(boolean);
 * <p>
 * This method is provided to allow a DateSelector pane to be constructed, configured
 * and then displayed.
 *
 * @param owner - the owning component or null
 * @param panel - the panel
 * @return the pop up panel
 */
public static PopupDateSelector createPopup(Component owner, final DateSelector panel)
    {
    PopupDateSelector pdts = new PopupDateSelector(owner, panel);

    return pdts;
    }

/**
 * Displays a modal dialog allowing the selection of a date
 *
 * @param owner - the owning component or null
 * @param title - The title to display in the dialog frame or null for no title 
 * @param now - the time now
 * @return the selected date or null if cancel pressed
 */
public static DateTime show(Component owner, String title, DateTime now)
    {
    DateSelector panel = new DateSelector(now, true);
    return show(owner, title, panel);
    }

/**
 * Displays a modal dialog allowing the selection of a date.
 * <p>
 * This method is provided to allow a DateSelector pane to be constructed, configured
 * and then displayed.
 *
 * @param owner - the owning component or null
 * @param title - The title to display in the dialog frame or null for no title 
 * @param panel - the panel
 * @return the selected date or null if cancel pressed
 */
public static DateTime show(Component owner, String title, DateSelector panel)
    {
    JDialog dlg = AbstractDateSelector.createDialog(owner, title, panel); 
    dlg.setVisible(true);

    return panel.getSelectedDate();
    }

/**
 * Construct a DateSelector for the default locale
 *
 * @param now - the initial date
 * @param includeOkCancel - true to include a panel with OK and Cancel buttons
 */
public DateSelector(DateTime now, boolean includeOkCancel)
    {
    this(now, includeOkCancel, null);
    }

/**
 * Construct a DateSelector
 *
 * @param now - the initial date
 * @param includeOkCancel - true to include a panel with OK and Cancel buttons
 * @param loc - the locale to use
 */
public DateSelector(DateTime now, boolean includeOkCancel, Locale loc)
    {
    super(now, now, includeOkCancel, loc);
    
    initKeyActions();
    }

/**
 * Gets the selected date
 *
 * @return the selected date or null if cancel was pressed
 */
public DateTime getSelectedDate()
    {
    if ( baseDate == null )
        return null;

    return baseDate.toDateTime(chronology);
    }

/**
 * Sets the selected date
 *
 * @param date - the selected date
 */
public void setSelectedDate(DateTime date)
    {
    if ( date == null )
        return;

    baseDate = date.toMutableDateTime(chronology);
    movingDate = baseDate;
    
    super.updateOnSettingSelectedDate();
    }


/**
 * The action to run on hovering over a registered component
 *
 * @author A.G.Docherty
 * @date 28 Jan 2010
 *
 */
private static class HoverAction implements Callable 
    {
    private MutableDateTime date;
    private ActionListener listener;
    private Component owner;

    HoverAction(Component o, MutableDateTime dt, ActionListener l)
        {
        owner = o;
        date = dt;
        listener = l;
        }
    
    @Override
    public Object call()
        {
        // create the date selector and show it
        PopupDateSelector dlg = DateSelector.createPopup(owner, date.toDateTime(date.getChronology()));
    
        dlg.addActionListener(listener);
        dlg.setVisible(true);
        
        return dlg;
        }
    }

/**
 * The Popup panel for displaying this panel
 *
 * @author A.G.Docherty
 * @date 20 Jan 2010
 *
 */
/**
 * The Popup panel for displaying this panel
 *
 * @author A.G.Docherty
 * @date 20 Jan 2010
 *
 */
public static class PopupDateSelector extends AbstractPopupSelector<DateSelector, DateTime> 
    {
    private static final long serialVersionUID = -5818592164102830422L;
    
    /**
     * Creates a PopupDateTimeSelector for this component which displays this panel
     *
     * @param parent - the component that triggered the popup
     * @param pane - the selector panel to display
     */
    public PopupDateSelector(Component parent, DateSelector pane)
        {
        super(parent, pane);
        }
    
    /**
     * Gets the DateSelector panel
     *
     * @return the panel
     */
    @Override
    public DateSelector getSelector() { return panel; }
    
    /**
     * Gets the selected date from the DateSelector
     * This is a convenience method that calls through to the DateSelector panel
     *
     * @return the selected date
     */
    @Override
    public DateTime getSelectedDate() { return panel.getSelectedDate(); }
    }
}
