/** 
 * Copyright (c) 2010 Keang Ltd
 * 
 * This file is part of the Date Selector package.
 *
 * The Date Selector package is free software: you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as published 
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * The Date Selector package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with the Date Selector package.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.co.keang.date;

import java.awt.Component;
import java.awt.event.ActionListener;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.util.Locale;
import java.util.concurrent.Callable;
import javax.swing.InputMap;
import javax.swing.JDialog;
import javax.swing.KeyStroke;
import org.joda.time.DateMidnight;
import org.joda.time.DateTime;
import org.joda.time.Interval;
import org.joda.time.MutableInterval;


/**
 * A Date Interval selection GUI.<br>
 * Displays a calendar style view of the days in a month. Two dates are involved in selecting an
 * interval, the base date is selected by selecting any day whilst the shift key is not pressed
 * and the second date is selected by selecting any day whilst the shift key is pressed. The month and year can be easily
 * changed via drop down controls and/or by month increment buttons and the day by
 * clicking on a day button.
 * <p>
 * If todays date is visible ie the panel is showing the todays month and year then the text
 * on the button with todays day is coloured RED.
 * <p>
 * As well as being able to select a date using the mouse the following key-strokes are recognised:<br>
 * LEFT: move the base date back 1 day<br>
 * RIGHT: move the base date forward one day<br>
 * UP: move the base date back 1 week<br>
 * DOWN: move the base date forward 1 week<br>
 * PAGE_UP: move the base date back 1 month<br>
 * PAGE_DOWN: move the base date forward 1 month<br>
 * HOME: move the base date back 1 year<br>
 * END: move the base date forward 1 year<br>
 * SHIFT+LEFT: move the second date back 1 day<br>
 * SHIFT+RIGHT: move the second date forward one day<br>
 * SHIFT+UP: move the second date back 1 week<br>
 * SHIFT+DOWN: move the second date forward 1 week<br>
 * SHIFT+PAGE_UP: move the second date back 1 month<br>
 * SHIFT+PAGE_DOWN: move the second date forward 1 month<br>
 * SHIFT+HOME: move the second date back 1 year<br>
 * SHIFT+END: move the second date forward 1 year<br>
 * ESC: cancel any selected date and hide the panel<br>
 * ENTER: retain the selected date and hide the panel<br>
 * <p>
 * Listeners can be added to receive notification of a change of selected date and/or the selection of
 * a day button.
 * <p>
 * Factory methods are provided to:<br>
 * 1. Show the selector in a JDialog optionally with OK/Cancel buttons<br>
 * 2. Show the selector in a Popup style panel<br>
 * 3. Attach a popup-on-hover feature to a component to popup a date selector if the mouse hovers over
 *    the component.<br>
 * <p>
 * Alternatively the panel can be constructed directly and added to an existing GUI.
 *
 *
 * @author A.G.Docherty
 * @date 14 Jan 2010
 *
 */
public class IntervalSelector extends AbstractDateSelector<Interval>
{
private static final long serialVersionUID = -2408963551442047759L;

/**
 * Adds a pop up listener to this component. When the mouse hovers over the component
 * an IntervalSelector popup will display.
 * <p>
 * The panel will close when:
 * 1. a day button is selected
 * 2. the mouse moves out of the panel
 * 3. the ESCAPE key is pressed
 * 4. the ENTER key is pressed
 * <p>
 * The given listener will be called only if the panel closes due to a day being selected or
 * the ENTER key is pressed
 *
 * @param owner - the component to add the hover popup listener to
 * @param dt - the date that will display when the popup pops up. This objects date can be dynamically 
 *             changed by the calling code to change the date that displays on popup
 * @param l - a listener that will be called when a day button is selected
 */
public static void setPopupOnHoverForComponent(Component owner, MutableInterval dt, ActionListener l)
    {
    AbstractDateSelector.setPopupOnHoverForComponent(owner, new HoverAction(owner, dt, l));
    }

/**
 * Removes the Popup on hover functionality from this component
 *
 * @param owner - the component
 */
public static void removePopupOnHoverFromComponent(Component owner)
    {
    AbstractDateSelector.removePopupOnHoverFromComponent(owner);
    }

/**
 * Creates a popup panel allowing the selection of a date interval. The panel is an undecorated
 * JDialog with OK and cancel buttons. To show the panel use setVisible(true);
 *
 * @param owner - the owning component or null
 * @param interval - the time now
 * @return the popup panel
 */
public static PopupIntervalSelector createPopup(Component owner, Interval interval)
    {
    IntervalSelector panel = new IntervalSelector(interval, true);
    return createPopup(owner, panel);
    }

/**
 * Creates a popup panel allowing the selection of a date interval. The panel is an undecorated
 * JDialog. To show/hide the panel use setVisible(boolean);
 * <p>
 * This method is provided to allow an IntervalSelector pane to be constructed, configured
 * and then displayed.
 *
 * @param owner - the owning component or null
 * @param panel - the panel
 * @return the pop up panel
 */
public static PopupIntervalSelector createPopup(Component owner, final IntervalSelector panel)
    {
    PopupIntervalSelector pdts = new PopupIntervalSelector(owner, panel);

    return pdts;
    }

/**
 * Displays a modal dialog allowing the selection of a date interval.
 *
 * @param owner - the owning component or null
 * @param title - The title to display in the dialog frame or null for no title 
 * @param now - the time now
 * @return the selected date or null if cancel pressed
 */
public static Interval show(Component owner, String title, Interval now)
    {
    IntervalSelector panel = new IntervalSelector(now, true);
    return show(owner, title, panel);
    }

/**
 * Displays a modal dialog allowing the selection of a date interval.
 * <p>
 * This method is provided to allow an IntervalSelector pane to be constructed, configured
 * and then displayed.
 *
 * @param owner - the owning component or null
 * @param title - The title to display in the dialog frame or null for no title 
 * @param panel - the panel
 * @return the selected date or null if cancel pressed
 */
public static Interval show(Component owner, String title, IntervalSelector panel)
    {
    JDialog dlg = AbstractDateSelector.createDialog(owner, title, panel); 
    dlg.setVisible(true);

    return panel.getSelectedDate();
    }

/**
 * Construct an IntervalSelector.
 * The base date is assumed to be the start date of the interval. 
 * <p>
 * An interval is half inclusive ie the start date is inclusive and the end date is exclusive.
 *
 * @param interval - the initial date interval
 * @param includeOkCancel - true to include a panel with OK and Cancel buttons
 */
public IntervalSelector(Interval interval, boolean includeOkCancel)
    {
    this(interval.getStart(), interval.getEnd().minusDays(1), includeOkCancel);
    }

/**
 * Construct an IntervalSelector.
 * The base date is assumed to be the start date of the interval.
 *
 * @param interval - the initial date interval
 * @param includeOkCancel - true to include a panel with OK and Cancel buttons
 */
public IntervalSelector(Interval interval, boolean includeOkCancel, Locale loc)
    {
    this(interval.getStart(), interval.getEnd().minusDays(1), includeOkCancel, loc);
    }

/**
 * Construct an IntervalSelector.
 * The base date is assumed to be the start date of the interval.
 * <p>
 * These values are both inclusive
 * 
 * @param base - the base date
 * @param end - the end date, this may be before or after the base date 
 * @param includeOkCancel - true to include a panel with OK and Cancel buttons
 */
public IntervalSelector(DateTime base, DateTime end, boolean includeOkCancel)
    {
    this(base, end, includeOkCancel, null);
    }

/**
 * Construct an IntervalSelector.
 * The base date is assumed to be the start date of the interval.
 *
 * @param base - the base date
 * @param end - the end date, this may be before or after the base date 
 * @param includeOkCancel - true to include a panel with OK and Cancel buttons
 */
public IntervalSelector(DateTime base, DateTime end, boolean includeOkCancel, Locale loc)
    {
    super(base, end, includeOkCancel, loc);

    initKeyActions();
    }

/**
 * init the key strokes that have meaning for this panel
 */
protected void initKeyActions()
    {
    super.initKeyActions();
    
    // add the day move key strokes
    InputMap imap = this.getInputMap(WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
    imap.put(KeyStroke.getKeyStroke(KeyEvent.VK_LEFT, InputEvent.SHIFT_DOWN_MASK), ACTION_NAME_DAY_DEC);
    imap.put(KeyStroke.getKeyStroke(KeyEvent.VK_RIGHT, InputEvent.SHIFT_DOWN_MASK), ACTION_NAME_DAY_INC);
    imap.put(KeyStroke.getKeyStroke(KeyEvent.VK_UP, InputEvent.SHIFT_DOWN_MASK), ACTION_NAME_WEEK_DEC);
    imap.put(KeyStroke.getKeyStroke(KeyEvent.VK_DOWN, InputEvent.SHIFT_DOWN_MASK), ACTION_NAME_WEEK_INC);

    // add the month move key strokes
    imap.put(KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_UP, InputEvent.SHIFT_DOWN_MASK), ACTION_NAME_MONTH_DEC);
    imap.put(KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_DOWN, InputEvent.SHIFT_DOWN_MASK), ACTION_NAME_MONTH_INC);

    // add the year move key strokes
    imap.put(KeyStroke.getKeyStroke(KeyEvent.VK_HOME, InputEvent.SHIFT_DOWN_MASK), ACTION_NAME_YEAR_DEC);
    imap.put(KeyStroke.getKeyStroke(KeyEvent.VK_END, InputEvent.SHIFT_DOWN_MASK), ACTION_NAME_YEAR_INC);
    }

/**
 * Gets the selected date interval.
 * <p>
 * The returned interval is half inclusive ie the start date is inclusive and the end
 * date is exclusive.
 *
 * @return the selected date or null if cancel was pressed
 */
public Interval getSelectedDate()
    {
    if ( baseDate == null )
        return null;

    DateMidnight bm = baseDate.toDateTime(chronology).toDateMidnight();
    DateMidnight mm = movingDate.toDateTime(chronology).toDateMidnight();
    
    if ( bm.isBefore(mm) )
        return new Interval(bm, mm.plusDays(1));
    else
        return new Interval(mm, bm.plusDays(1));
    }

/**
 * Sets the selected date interval. 
 * The base date is assumed to be the start date of the interval.
 * <p>
 * The interval is half inclusive ie the start date is inclusive and the end
 * date is exclusive.
 *
 * @param date - the selected date
 */
public void setSelectedDate(Interval date)
    {
    if ( date == null )
        throw new IllegalArgumentException("The date cannot be null");

    baseDate = date.getStart().toDateMidnight().toMutableDateTime(chronology);
    movingDate = date.getEnd().toDateMidnight().minusDays(1).toMutableDateTime(chronology);

    super.updateOnSettingSelectedDate();
    }

/**
 * Sets the selected date interval. 
 * <p>
 * The dates are fully inclusive
 * 
 * @param base - the base date
 * @param end - the end date, this may be before or after the base date 
 */
public void setSelectedDate(DateTime base, DateTime end)
    {
    if ( base == null )
        throw new IllegalArgumentException("The base date cannot be null");
    
    if ( end == null )
        throw new IllegalArgumentException("The end date cannot be null");

    baseDate = base.toDateMidnight().toMutableDateTime(chronology);
    movingDate = end.toDateMidnight().toMutableDateTime(chronology);

    super.updateOnSettingSelectedDate();
    }

/**
 * The action to run on hovering over a registered component
 *
 * @author A.G.Docherty
 * @date 28 Jan 2010
 *
 */
private static class HoverAction implements Callable 
    {
    private MutableInterval date;
    private ActionListener listener;
    private Component owner;

    HoverAction(Component o, MutableInterval dt, ActionListener l)
        {
        owner = o;
        date = dt;
        listener = l;
        }
    
    @Override
    public Object call()
        {
        // create the date selector and show it
        PopupIntervalSelector dlg = IntervalSelector.createPopup(owner, date.toInterval());
    
        dlg.addActionListener(listener);
        dlg.setVisible(true);
        
        return dlg;
        }
    }

/**
 * The Popup panel for displaying this Interval Selector
 *
 * @author A.G.Docherty
 * @date 20 Jan 2010
 *
 */
public static class PopupIntervalSelector extends AbstractPopupSelector<IntervalSelector, Interval> 
    {
    private static final long serialVersionUID = -5812051277389233706L;

    /**
     * Creates a PopupDateTimeSelector for this component which displays this panel
     *
     * @param parent - the component that triggered the popup
     * @param pane - the selector panel to display
     */
    public PopupIntervalSelector(Component parent, IntervalSelector pane)
        {
        super(parent, pane);
        }
    
    /**
     * Gets the JDateTimeSelector panel
     *
     * @return the the panel
     */
    @Override
    public IntervalSelector getSelector() { return panel; }
    
    /**
     * Gets the selected date from the JDateTimeSelector
     * This is a convenience method that calls through to the JDateTimeSelector panel
     *
     * @return the date
     */
    @Override
    public Interval getSelectedDate() { return panel.getSelectedDate(); }
    }

}
